"""
Randomized tests of GiNaC / PyNaC
"""

###############################################################################
#   Sage: Open Source Mathematical Software
#       Copyright (C) 2008 William Stein <wstein@gmail.com>
#       Copyright (C) 2008 Burcin Erocal <burcin@erocal.org>
#  Distributed under the terms of the GNU General Public License (GPL),
#  version 2 or any later version.  The full text of the GPL is available at:
#                  https://www.gnu.org/licenses/
###############################################################################


from sage.misc.prandom import randint, random
import operator
from sage.rings.rational_field import QQ
from sage.symbolic.ring import SR
from sage.symbolic.expression import symbol_table, mixed_order
from sage.symbolic.constants import (pi, e, golden_ratio, log2, euler_gamma,
                                     catalan, khinchin, twinprime, mertens)
from sage.functions.hypergeometric import hypergeometric
from sage.functions.other import (cases, element_of)


##############################################################
#        Generate random expressions for doctests            #
##############################################################

def _mk_full_functions():
    r"""
    A simple function that returns a list of all Pynac functions of known
    arity, sorted by name.

    EXAMPLES::

        sage: from sage.symbolic.random_tests import _mk_full_functions
        sage: [f for (one,f,arity) in _mk_full_functions()] # random
        [Ei, abs, arccos, arccosh, arccot, arccoth, arccsc, arccsch,
        arcsec, arcsech, arcsin, arcsinh, arctan, arctan2, arctanh,
        arg, beta, binomial, ceil, conjugate, cos, cosh, cot, coth,
        csc, csch, dickman_rho, dilog, dirac_delta, elliptic_e,
        elliptic_ec, elliptic_eu, elliptic_f, elliptic_kc,
        elliptic_pi, erf, exp, factorial, floor, heaviside, imag_part,
        integrate, kronecker_delta, log, polylog, real_part, sec,
        sech, sgn, sin, sinh, tan, tanh, unit_step, zeta, zetaderiv]

    Note that this doctest will produce different output whenever a
    symbolic function is added or removed.
    """
    excluded = [hypergeometric, cases, element_of]
    items = sorted(symbol_table['functions'].items())
    return [(1.0, f, f.number_of_arguments())
            for (name, f) in items
            if hasattr(f, 'number_of_arguments') and
            f.number_of_arguments() > 0 and
            f not in excluded]


# For creating simple expressions

fast_binary = [(0.4, operator.add), (0.1, operator.sub), (0.5, operator.mul)]
fast_unary = [(0.8, operator.neg), (0.2, operator.abs)]
fast_nodes = [(0.9, fast_binary, 2), (0.1, fast_unary, 1)]

# For creating expressions with the full power of Pynac's simple expression
# subset (with no quantifiers/operators; that is, no derivatives, integrals,
# etc.)
full_binary = [(0.3, operator.add), (0.1, operator.sub), (0.3, operator.mul), (0.2, operator.truediv), (0.1, operator.pow)]
full_unary = [(0.8, operator.neg), (0.2, operator.inv)]
full_functions = _mk_full_functions()
full_nullary = [(1.0, c) for c in [pi, e]] + [(0.05, c) for c in
        [golden_ratio, log2, euler_gamma, catalan, khinchin, twinprime,
            mertens]]
full_internal = [(0.6, full_binary, 2), (0.2, full_unary, 1),
        (0.2, full_functions)]


def normalize_prob_list(pl, extra=()):
    r"""
    INPUT:

    - ``pl`` -- list of tuples, where the first element of each tuple is
      a floating-point number (representing a relative probability).  The
      second element of each tuple may be a list or any other kind of object.

    - ``extra`` -- tuple which is to be appended to every tuple in ``pl``

    This function takes such a list of tuples (a "probability list") and
    normalizes the probabilities so that they sum to one.  If any of the
    values are lists, then those lists are first normalized; then
    the probabilities in the list are multiplied by the main probability
    and the sublist is merged with the main list.

    For example, suppose we want to select between group A and group B with
    50% probability each.  Then within group A, we select A1 or A2 with 50%
    probability each (so the overall probability of selecting A1 is 25%);
    and within group B, we select B1, B2, or B3 with probabilities in
    a 1:2:2 ratio.

    EXAMPLES::

        sage: from sage.symbolic.random_tests import *
        sage: A = [(0.5, 'A1'), (0.5, 'A2')]
        sage: B = [(1, 'B1'), (2, 'B2'), (2, 'B3')]
        sage: top = [(50, A, 'Group A'), (50, B, 'Group B')]
        sage: normalize_prob_list(top)
        [(0.250000000000000, 'A1', 'Group A'), (0.250000000000000, 'A2', 'Group A'), (0.1, 'B1', 'Group B'), (0.2, 'B2', 'Group B'), (0.2, 'B3', 'Group B')]
    """
    if len(pl) == 0:
        return pl
    result = []
    total = sum(float(p[0]) for p in pl)
    for p in pl:
        prob = p[0]
        val = p[1]
        if len(p) > 2:
            p_extra = p[2:]
        else:
            p_extra = extra
        if isinstance(val, list):
            norm_val = normalize_prob_list(val, extra=p_extra)
            result.extend(((prob / total) * np[0], np[1]) + np[2:]
                          for np in norm_val)
        else:
            result.append(((prob / total), val) + p_extra)
    return result


def choose_from_prob_list(lst):
    r"""
    INPUT:

    - ``lst`` -- list of tuples, where the first element of each tuple
      is a nonnegative float (a probability), and the probabilities sum
      to one

    OUTPUT:

    A tuple randomly selected from the list according to the given
    probabilities.

    EXAMPLES::

        sage: from sage.symbolic.random_tests import *
        sage: v = [(0.1, False), (0.9, True)]
        sage: choose_from_prob_list(v)  # random
        (0.900000000000000, True)
        sage: true_count = 0
        sage: total_count = 0
        sage: def more_samples():
        ....:     global true_count, total_count
        ....:     for _ in range(10000):
        ....:         total_count += 1.0
        ....:         if choose_from_prob_list(v)[1]:
        ....:             true_count += 1.0
        sage: more_samples()
        sage: while abs(true_count/total_count - 0.9) > 0.01:
        ....:     more_samples()
    """
    r = random()
    for i in range(len(lst) - 1):
        if r < lst[i][0]:
            return lst[i]
        r -= lst[i][0]
    return lst[-1]


def random_integer_vector(n, length):
    r"""
    Give a random list of length *length*, consisting of nonnegative
    integers that sum to *n*.

    This is an approximation to IntegerVectors(n, length).random_element().
    That gives values uniformly at random, but might be slow; this
    routine is not uniform, but should always be fast.

    (This routine is uniform if ``length`` is 1 or 2; for longer vectors,
    we prefer approximately balanced vectors, where all the values
    are around `n/{length}`.)

    EXAMPLES::

        sage: from sage.symbolic.random_tests import *
        sage: a = random_integer_vector(100, 2); a  # random
        [11, 89]
        sage: len(a)
        2
        sage: sum(a)
        100

        sage: b = random_integer_vector(10000, 20)
        sage: len(b)
        20
        sage: sum(b)
        10000

    The routine is uniform if ``length`` is 2::

        sage: true_count = 0
        sage: total_count = 0
        sage: def more_samples():
        ....:     global true_count, total_count
        ....:     for _ in range(1000):
        ....:         total_count += 1.0
        ....:         if a == random_integer_vector(100, 2):
        ....:             true_count += 1.0
        sage: more_samples()
        sage: while abs(true_count/total_count - 0.01) > 0.01:
        ....:     more_samples()
    """
    if length == 0:
        return []
    if length == 1:
        return [n]
    if length == 2:
        v = randint(0, n)
        return [v, n - v]
    v = randint(0, 2 * n // length)
    return [v] + random_integer_vector(n - v, length - 1)


def random_expr_helper(n_nodes, internal, leaves, verbose):
    r"""
    Produce a random symbolic expression of size *n_nodes* (or slightly
    larger).  Internal nodes are selected from the *internal* probability
    list; leaves are selected from *leaves*.  If *verbose* is True,
    then a message is printed before creating an internal node.

    EXAMPLES::

        sage: from sage.symbolic.random_tests import *
        sage: a = random_expr_helper(9, [(0.5, operator.add, 2),
        ....:     (0.5, operator.neg, 1)], [(0.5, 1), (0.5, x)], True)
        About to apply <built-in function ...

    In small cases we will see all cases quickly::

        sage: def next_expr():
        ....:     return random_expr_helper(
        ....:         6, [(0.5, operator.add, 2), (0.5, operator.neg, 1)],
        ....:         [(0.5, 1), (0.5, x)], False)
        sage: all_exprs = set()
        sage: for a in range(-4, 5):
        ....:     for b in range(-4+abs(a), 5-abs(a)):
        ....:         if a % 2 and abs(a) + abs(b) == 4 and sign(a) != sign(b):
        ....:             continue
        ....:         all_exprs.add(a*x + b)
        sage: our_exprs = set()
        sage: while our_exprs != all_exprs:
        ....:    our_exprs.add(next_expr())
    """
    if n_nodes == 1:
        return choose_from_prob_list(leaves)[1]
    r = choose_from_prob_list(internal)
    n_nodes -= 1
    n_children = r[2]
    n_spare_nodes = n_nodes - n_children
    n_spare_nodes = max(0, n_spare_nodes)
    nodes_per_child = random_integer_vector(n_spare_nodes, n_children)
    children = [random_expr_helper(n + 1, internal, leaves, verbose)
                for n in nodes_per_child]
    if verbose:
        print("About to apply %r to %r" % (r[1], children))
    return r[1](*children)


def random_expr(size, nvars=1, ncoeffs=None, var_frac=0.5,
                internal=full_internal,
                nullary=full_nullary, nullary_frac=0.2,
                coeff_generator=QQ.random_element, verbose=False):
    r"""
    Produce a random symbolic expression of the given size.  By
    default, the expression involves (at most) one variable, an arbitrary
    number of coefficients, and all of the symbolic functions and constants
    (from the probability lists full_internal and full_nullary).  It is
    possible to adjust the ratio of leaves between symbolic constants,
    variables, and coefficients (var_frac gives the fraction of variables,
    and nullary_frac the fraction of symbolic constants; the remaining
    leaves are coefficients).

    The actual mix of symbolic constants and internal nodes can be modified
    by specifying different probability lists.

    To use a different type for coefficients, you can specify
    coeff_generator, which should be a function that will return
    a random coefficient every time it is called.

    This function will often raise an error because it tries to create
    an erroneous expression (such as a division by zero).

    EXAMPLES::

        sage: from sage.symbolic.random_tests import *
        sage: some_functions = [arcsinh, arctan, arctan2, arctanh,
        ....: arg, beta, binomial, ceil, conjugate, cos, cosh, cot, coth,
        ....: elliptic_pi, erf, exp, factorial, floor, heaviside, imag_part,
        ....: sech, sgn, sin, sinh, tan, tanh, unit_step, zeta, zetaderiv]
        sage: my_internal = [(0.6, full_binary, 2), (0.2, full_unary, 1),
        ....: (0.2, [(1.0,f,f.number_of_arguments()) for f in some_functions])]
        sage: set_random_seed(1)
        sage: random_expr(50, nvars=3, internal=my_internal,  # not tested  # known bug
        ....:             coeff_generator=CDF.random_element)
        (v1^(0.9713408427702117 + 0.195868299334218*I)/cot(-pi + v1^2 + v3) + tan(arctan(v2 + arctan2(-0.35859061674557324 + 0.9407509502498164*I, v3) - 0.8419115504372718 + 0.30375717982404615*I) + arctan2((0.2275357305882964 - 0.8258002386106038*I)/factorial(v2), -v3 - 0.7604559947718565 - 0.5543672548552057*I) + ceil(1/arctan2(v1, v1))))/v2
        sage: random_expr(5, verbose=True)  # not tested  # known bug
        About to apply <built-in function inv> to [31]
        About to apply sgn to [v1]
        About to apply <built-in function add> to [1/31, sgn(v1)]
        sgn(v1) + 1/31
    """
    vars = [(1.0, SR.var('v%d' % (n + 1))) for n in range(nvars)]
    if ncoeffs is None:
        ncoeffs = size
    coeffs = [(1.0, coeff_generator()) for _ in range(ncoeffs)]
    leaves = [(var_frac, vars), (1.0 - var_frac - nullary_frac, coeffs), (nullary_frac, nullary)]
    leaves = normalize_prob_list(leaves)

    internal = normalize_prob_list(internal)

    return random_expr_helper(size, internal, leaves, verbose)


#####################################
#   Test the ordering of operands   #
#####################################

def assert_strict_weak_order(a, b, c, cmp_func):
    r"""
    Check that ``cmp_func`` is a strict weak order on the elements a,b,c.

    A strict weak order is a binary relation ``<`` such that

    * For all `x`, it is not the case that `x < x` (irreflexivity).

    * For all `x\not=y`, if `x < y` then it is not the case that `y <
      x` (asymmetry).

    * For all `x`, `y`, and `z`, if `x < y` and `y < z` then `x < z`
      (transitivity).

    * For all `x`, `y`, and `z`, if x is incomparable with `y`, and
      `y` is incomparable with `z`, then `x` is incomparable with `z`
      (transitivity of incomparability).

    INPUT:

    - ``a``, ``b``, ``c`` -- anything that can be compared by ``cmp_func``

    - ``cmp_func`` -- function of two arguments that returns their
      comparison (i.e. either ``True`` or ``False``)

    OUTPUT:

    Nothing. Raises a :exc:`ValueError` if ``cmp_func``
    is not a strict weak order on the three given elements.

    REFERENCES:

    :wikipedia:`Strict_weak_ordering`

    EXAMPLES:

    The usual ordering of integers is a strict weak order::

        sage: from sage.symbolic.random_tests import assert_strict_weak_order
        sage: a, b, c = [randint(-10, 10) for i in range(3)]
        sage: assert_strict_weak_order(a, b, c, lambda x, y: x < y)

        sage: x = [-SR(oo), SR(0), SR(oo)]
        sage: cmp_M = matrix(3, 3, 0)
        sage: for i in range(3):
        ....:     for j in range(3):
        ....:         if x[i] < x[j]:
        ....:             cmp_M[i, j] = -1
        ....:         elif x[i] > x[j]:
        ....:             cmp_M[i, j] = 1
        sage: cmp_M
        [ 0 -1 -1]
        [ 1  0 -1]
        [ 1  1  0]
    """
    from sage.matrix.constructor import matrix
    from sage.combinat.permutation import Permutations
    x = (a, b, c)

    cmp_M = matrix(3, 3)
    for i in range(3):
        for j in range(3):
            cmp_M[i, j] = (cmp_func(x[i], x[j]) == 1)   # or -1, doesn't matter

    msg = 'the binary relation failed to be a strict weak order on the elements \n'
    msg += ' a = {}\n b = {}\n c = {}\n'.format(a, b, c)
    msg += str(cmp_M)

    for i in range(3):
        # irreflexivity
        if cmp_M[i, i]:
            raise ValueError(msg)

        # asymmetric
        for j in range(i):
            if cmp_M[i, j] and cmp_M[j, i]:
                raise ValueError(msg)

    def incomparable(i, j):
        return not (cmp_M[i, j] or cmp_M[j, i])

    for i, j, k in Permutations([0, 1, 2]):
        # transitivity
        if cmp_M[i, j] and cmp_M[j, k] and not cmp_M[i, k]:
            raise ValueError(msg)

        # transitivity of incomparability
        if (incomparable(i, j) and incomparable(j, k) and
                not incomparable(i, k)):
            raise ValueError(msg)


def check_symbolic_expression_order(repetitions=100):
    r"""
    Test whether the comparison of random symbolic expressions
    satisfies the strict weak order axioms.

    This is important because the C++ extension class uses
    ``std::sort()`` which requires a strict weak order. See also
    :issue:`9880`.

    EXAMPLES::

        sage: from sage.symbolic.random_tests import check_symbolic_expression_order
        sage: check_symbolic_expression_order(500)  # long time
    """
    rnd_length = 50
    nvars = 10
    ncoeffs = 10
    var_frac = 0.5
    nullary_frac = 0.05

    def coeff_generator():
        return randint(-100, 100) / randint(1, 100)

    def make_random_expr():
        while True:
            try:
                return random_expr(
                    rnd_length, nvars=nvars, ncoeffs=ncoeffs, var_frac=var_frac,
                    nullary_frac=nullary_frac, coeff_generator=coeff_generator,
                    internal=fast_nodes)
            except (ZeroDivisionError, ValueError):
                pass

    for rep in range(repetitions):
        a = make_random_expr()
        b = make_random_expr()
        c = make_random_expr()
        assert_strict_weak_order(a, b, c, mixed_order)
        assert_strict_weak_order(a, b, c, lambda x, y: x._cmp_add(y))
        assert_strict_weak_order(a, b, c, lambda x, y: x._cmp_mul(y))



# sage.doctest: optional regina
r"""
Interface to Regina

`Regina <https://regina-normal.github.io/>`__ is an open source software for
low-dimensional topology. From the home-page:

.. NOTE::

    Regina is a software package for 3-manifold and 4-manifold
    topologists, with a focus on triangulations, knots and links,
    normal surfaces, and angle structures.

    For 3-manifolds, it includes high-level tasks such as 3-sphere
    recognition, connected sum decomposition and Hakenness testing,
    comes with a rich database of census manifolds, and incorporates
    the SnapPea kernel for working with hyperbolic manifolds.
    For 4-manifolds, it offers a range of combinatorial and algebraic
    tools, plus support for normal hypersurfaces. For knots and links,
    Regina can perform combinatorial manipulation, compute knot
    polynomials, and work with several import/export formats.

The Regina interface will only work if the optional Sage package Regina
is installed. The interface lets you send certain Sage objects to Regina,
run Regina functions, import certain Regina expressions to Sage,
or any combination of the above.

To send a Sage object ``sobj`` to Regina, call ``regina(sobj)``.
This exports the Sage object to Regina and returns a new Sage object
wrapping the Regina expression/variable, so that you can use the
Regina variable from within Sage. You can then call Regina
functions on the new object; for example::

    sage: F3 = FreeGroup(3)
    sage: F3r = regina(F3); F3r
    <regina.GroupPresentation: < a b c >>
    sage: (F3r.parent(), type(F3r))
    (Regina, <class 'sage.interfaces.regina.ReginaElement'>)
    sage: f = F3.an_element(); f
    x0*x1*x2
    sage: fr = regina(f); fr
    <regina.GroupExpression: g0 g1 g2>
    sage: rel = fr^2; rel
    <regina.GroupExpression: g0 g1 g2 g0 g1 g2>
    sage: (type(fr), type(rel))
    (<class 'sage.interfaces.regina.ReginaElement'>,
     <class 'sage.interfaces.regina.ReginaElement'>)
    sage: F3r.addRelation(rel); F3r
    <regina.GroupPresentation: < a b c | a b c a b c >>

In the above example the relations are added using Regina's
``addRelations`` method.

To see Regina's output you can simply print the Regina wrapper
object. However if you want to import Regina's output back to Sage,
call the Regina wrapper object's ``sage()`` method. This method returns
a native Sage object::

    sage: G3 = F3r.sage(); G3
    Finitely presented group < x0, x1, x2 | (x0*x1*x2)^2 >
    sage: type(G3)
    <class 'sage.groups.finitely_presented.FinitelyPresentedGroup_with_category'>
    sage: regina(G3) == F3r
    True
    sage: fr.sage() == f
    True
    sage: rel.sage() == f^2
    True

If you want to run a Regina function and don't already have the input
in the form of a Sage object, then it might be simpler to input a string
``expr`` to ``regina(expr)``. This string will be evaluated as if you had
typed it into Regina::

    sage: rL = regina("Link('dabcabcv-')"); rL
    <regina.Link: 3-crossing knot: +++ ( ^0 _1 ^2 _0 ^1 _2 )>

Alternatively, all constructors of Regina classes can be used directly as attributes
of the interface::

    sage: rL == regina.Link('dabcabcv-')
    True

Finally, if you just want to use a Regina command line from within
Sage, the IPython magic function ``%regina`` dumps you into an interactive
command-line Regina session. As long as you work in this environment the
prompt is ``regina:``. To finish the environment type ``CTRL+D``::

    sage: %regina                                 # not tested

    --> Switching to Regina <--

    regina: u = Link()
    None
    regina: u
    <regina.Link: Empty link>
    regina: type(u)
    <class 'regina.engine.Link'>
    regina: v = u.fromKnotSig('iabcdbefcdghaefghRsgF+m')
    None
    regina: v
    <regina.Link: 8-crossing knot: ++++--++ ( ^0 ^1 _2 ^3 _1 ^4 _5 ^2 _3 _6 ^7 _0 _4 ^5 ^6 _7 )>
    regina: type(v)
    <class 'regina.engine.Link'>
    regina: v.homfly()
    <regina.Laurent2: 2 x^-2 y^2 + 3 x^-2 - x^-4 y^4 - 3 x^-4 y^2 - 3 x^-4 + x^-6 y^2 + x^-6>
    regina: u.homfly()
    <regina.Laurent2: 0>
    regina: type(u)
    <class 'regina.engine.Link'>

    --> Exiting back to Sage <--

    sage:                                         # not tested


Complicated translations
------------------------

The ``robj.sage()`` method tries to convert a Regina object to a Sage
object. In many cases, it will just work. In particular, it should be able to
convert expressions entirely consisting of:

- numbers, i.e. integers, floats, complex numbers;
- functions and named constants also present in Sage, where:

    - Sage knows how to translate the function or constant's name from
      Regina's, or
    - the Sage name for the function or constant is trivially related to
      Regina's;

- symbolic variables whose names don't pathologically overlap with
  objects already defined in Sage.

This method will not work when Regina's output includes:

- strings;
- functions unknown to Sage;
- Regina functions with different parameters/parameter order to
  the Sage equivalent.


AUTHORS:

- Sebastian Oehms (2025): first version.
"""

##############################################################################
#       Copyright (C) 2025 Sebastian Oehms <seb.oehms@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
##############################################################################

from enum import Enum
from sage.interfaces.interface import Interface, InterfaceElement, InterfaceFunction, InterfaceFunctionElement
from sage.interfaces.tab_completion import ExtraTabCompletion
from sage.misc.instancedoc import instancedoc


class AlgorithmExt(Enum):
    r"""
    Enum to select algorithm choices.

    This extends the ``Algorithm`` class of Regina.
    """
    ALG_SIMPLIFY = 4
    ALG_WIRTINGER = 5
    ALG_USE_EXTERIOR = 6


class Regina(ExtraTabCompletion, Interface):
    r"""
    Interface to the Regina interpreter.

    EXAMPLES::

        sage: K = Knots().from_table(8, 21)
        sage: Kr = regina(K); Kr
        <regina.Link: 8-crossing knot: ----++-- ( _5 _0 ^1 _2 _3 ^6 _7 ^3 _4 ^5 _6 ^7 ^0 _1 ^2 ^4 )>
        sage: Kr.knotSig()
        'iabcdbefcdghaefghRsgF+m'

    More examples can be found in the module header.
    """
    def __init__(self):
        r"""
        Python constructor.

        TESTS::

            sage: TestSuite(regina).run(skip=['_test_pickling', '_test_category'])
        """
        Interface.__init__(self, name='regina')
        self._initialized = False  # done lazily
        self._regina_globals = None
        self._namespace = None

    def _lazy_init(self):
        r"""
        Initialize the Regina interpreter.

        Implemented according to R interface.

        EXAMPLES::

            sage: regina._lazy_init()
        """
        if not self._initialized:
            self._initialized = True
            self._start()

    def __reduce__(self):
        r"""
        Helper for pickling.

        EXAMPLES::

            sage: p = regina.Polynomial([-2, ~7])  # indirect doctest
            sage: loads(dumps(p)) == p
            True
        """
        return self.__class__, ()

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        EXAMPLES::

            sage: regina.an_element()      # indirect doctest
            2
        """
        return self(2)

    def _start(self):
        """
        Start up the Regina interpreter and sets the initial prompt and options.

        This is called the first time the Regina interface is actually used.

        EXAMPLES::

            sage: regina._start()
            sage: type(regina._namespace.Cyclotomic())
            <class 'regina.engine.Cyclotomic'>
        """
        if not self._regina_globals:
            from sage.features.interfaces import Regina
            Regina().module.require()
            import regina
            self._namespace = regina.engine
            d = self._namespace.__dict__
            # add extras to the fixed namespace
            d['AlgorithmExt'] = AlgorithmExt
            for e in AlgorithmExt:
                d[e.name] = e
            # set up the globals
            D = dict(d)
            D[self._namespace.__name__] = self._namespace
            D[self.name()] = regina
            self._regina_globals = D

    def _install_hints(self):
        """
        Hints for installing regina on your computer.

        EXAMPLES::

            sage: len(regina._install_hints())
            99
        """
        return """
In order to use the Regina interface you need to have the
optional Sage package Regina installed.
"""

    def _eval(self, code):
        """
        Evaluates a command inside the Regina interpreter and returns the output
        as a Regina result.

        EXAMPLES::

            sage: regina._eval("Link('iabcdbefcdghaefghRsgF+m')")
            <regina.Link: 8-crossing knot: ++++--++ ( ^0 ^1 _2 ^3 _1 ^4 _5 ^2 _3 _6 ^7 _0 _4 ^5 ^6 _7 )>
        """
        self._lazy_init()
        globs = self._regina_globals
        if type(code) is str and code.find('=') < 0:
            pre_bracket = code.split('(')[0]
            if pre_bracket in globs:
                val = globs[pre_bracket]
                nam = self._namespace.__name__
                if callable(val):
                    return eval('%s.%s' % (nam, code), globs)
                return val
            return eval(code, globs)
        return exec(code, globs)

    def eval(self, code, *args, **kwds):
        """
        Evaluates a command inside the Regina interpreter and returns the output
        in printable form.

        EXAMPLES::

            sage: regina.eval('1+1')
            '2'
        """
        return repr(self._eval(code))

    def get(self, var):
        """
        Get the value of the variable ``var``.

        EXAMPLES::

            sage: regina.get('Link')
            <class 'regina.engine.Link'>
        """
        return self._regina_globals[var]

    def set(self, var, value):
        """
        Set the variable ``var`` to the given ``value``.

        EXAMPLES::

            sage: regina.set('myLink', 'Link')
            sage: regina.get('myLink')
            <class 'regina.engine.Link'>
        """
        self._lazy_init()
        globs = self._regina_globals
        if not isinstance(value, str):
            globs[var] = value
            return
        try:
            val = self._eval(value)
            globs[var] = val
        except (NameError, AttributeError, KeyError):
            pass
        super().set(var, value)

    def _regina_object(self, x) -> bool:
        r"""
        Return ``True`` if ``x`` is an instance of a class
        from Regina's namespace.

        EXAMPLES::

            sage: L = regina.get('Link')
            sage: regina._regina_object(L)
            False
            sage: regina._regina_object(L())
            True
            sage: regina._regina_object([])
            False
        """
        self._lazy_init()
        d = self._namespace.__dict__
        cl = x.__class__
        cln = cl.__name__
        if cln in d:
            if cl == d[cln]:
                return True
        return False

    def _coerce_from_special_method(self, x):
        """
        Try to coerce to ``self`` by calling a special underscore method.

        This method is overloaded to record the Sage parent of ``x`` in the
        interface element.

        EXAMPLES::

            sage: R.<u, v> = LaurentPolynomialRing(ZZ)
            sage: p = u*~v^3 + 3*v*~u + 5*u - 7
            sage: rp = regina(p)   # indirect doctest
            sage: rp._sage_parent
            Multivariate Laurent Polynomial Ring in u, v over Integer Ring
        """
        res = super()._coerce_from_special_method(x)
        if hasattr(x, 'parent'):
            res._sage_parent = x.parent()
        return res

    def _coerce_impl(self, x, use_special=True):
        r"""
        Coerce pure Python types via corresponding Sage objects.

        This method is overloaded to add Python types from regina._namespace

        EXAMPLES::

            sage: L = regina.get('Link')
            sage: regina._coerce_impl(L())
            <regina.Link: Empty link>
        """
        if self._regina_object(x):
            return self(self._create(x))
        return super()._coerce_impl(x, use_special=use_special)

    def _convert_args_kwds(self, *args, **kwds):
        """
        Convert all of the ``args`` and ``kwds`` to instances of Regina
        classes.

        EXAMPLES::

            sage: a = [regina(i) for i in range(3)]
            sage: b = list(range(3))
            sage: L = regina.get('Link')
            sage: C = L.fromKnotSig('iabcdbefcdghaefghRsgF+m')
            sage: D = (3,7)
            sage: regina._convert_args_kwds(a, b, C=C, D=D)
            (([0, 1, 2], [0, 1, 2]),
            {'C': <regina.Link: 8-crossing knot: ++++--++ ( ^0 ^1 _2 ^3 _1 ^4 _5 ^2 _3 _6 ^7 _0 _4 ^5 ^6 _7 )>,
             'D': (3, 7)})
        """
        def convert_arg(arg):
            if isinstance(arg, InterfaceElement) and arg.parent() is self:
                return arg._inst
            if isinstance(arg, (list, tuple)):
                return type(arg)([convert_arg(i) for i in arg])
            if hasattr(arg, '_regina_'):
                reg = arg._regina_(self)
                return convert_arg(reg)
            return arg

        if args:
            args = list(args)
            for i, arg in enumerate(args):
                args[i] = convert_arg(arg)
        if kwds:
            for key, value in kwds.items():
                kwds[key] = convert_arg(value)
        return tuple(args), kwds

    def _function_call(self, name, *args, **kwds):
        r"""
        Perform a function call.

        EXAMPLES::

            sage: regina._function_call(regina.PolynomialRational, (-3, 5/3))
            <regina.PolynomialRational: 5/3 x - 3>
        """
        args, kwds = self._convert_args_kwds(*args, **kwds)
        if len(args) == 0:
            if len(kwds) == 0:
                res = name()
            else:
                res = name(**kwds)
        elif len(args) == 1:
            if len(kwds) == 0:
                res = name(args[0])
            else:
                res = name(args[0], **kwds)
        else:
            if len(kwds) == 0:
                res = name(*args)
            else:
                res = name(*args, **kwds)

        # read back new values of the arguments and keywords
        def read_back(arg):
            if isinstance(arg, self._object_class()):
                self.set(arg._name, arg._inst)
        for arg in args:
            read_back(arg)
        for val in kwds.values():
            read_back(val)
        if res is not None:
            if self._regina_object(res):
                new = res.__class__(res)  # this is the way to get a copy of a Regina object
                return self(new)
            return res

    def _equality_symbol(self):
        r"""
        EXAMPLES::

            sage: regina._equality_symbol()
            '=='
        """
        return '=='

    def _object_class(self):
        r"""
        Return the element class of this parent.
        This is used in the interface class.

        EXAMPLES::

            sage: regina._object_class()
            <class 'sage.interfaces.regina.ReginaElement'>
        """
        return ReginaElement

    def help(self, cmd, long=False):
        r"""
        Return the Regina documentation of the given command.

        EXAMPLES::

            sage: regina.help('AbelianGroup')
            Represents a finitely generated abelian group.
            <BLANKLINE>
            The torsion elements of the group are stored in terms of their
            invariant factors. For instance, Z_2+Z_3 will appear as Z_6, and
            Z_2+Z_2+Z_3 will appear as Z_2+Z_6.
            <BLANKLINE>
            In general the factors will appear as Z_*d0*+...+Z_*dn*, where the
            invariant factors *di* are all greater than 1 and satisfy
            *d0*|*d1*|...|*dn*. Note that this representation is unique.
            <BLANKLINE>
            This class implements C++ move semantics and adheres to the C++
            Swappable requirement. It is designed to avoid deep copies wherever
            possible, even when passing or returning objects by value.
        """
        self._lazy_init()
        dic = self._namespace.__dict__
        if cmd in dic:
            return print(dic[cmd].__doc__)
        raise NotImplementedError('no documentation available for %s' % cmd)

    def _tab_completion(self):
        r"""
        Return a list of all classes available through the interface.

        .. NOTE::

           Currently returns all keys of the namspace dictionary.

        EXAMPLES::

            sage: 'AbelianGroup' in regina._tab_completion()
            True
        """
        self._lazy_init()
        return list(self._namespace.__dict__)

    def __getattr__(self, attrname):
        r"""

        EXAMPLES::

            sage: type(regina.AbelianGroup)
            <class 'sage.interfaces.regina.ReginaFunction'>
            sage: regina.AbelianGroup._name
            <class 'regina.engine.AbelianGroup'>
        """
        if attrname[:1] == "_":
            raise AttributeError
        self._lazy_init()
        try:
            attr = self._namespace.__dict__[attrname]
        except KeyError:
            raise AttributeError
        if callable(attr):
            return ReginaFunction(self, self._namespace.__dict__[attrname])
        res = self(self._create(attr))
        res._inst = attr
        return res


@instancedoc
class ReginaElement(ExtraTabCompletion, InterfaceElement):
    r"""
    Element class of the Regina interface.

    Its instances are usually constructed via the instance call of its parent.
    It wrapes the Regina library for this object. In a session Regina methods
    can be obtained using tab completion.

    EXAMPLES::

        sage: b = BraidGroup(3)((1,2,-1))
        sage: re = regina(b); re
        <regina.GroupExpression: g0 g1 g0^-1>
        sage: type(re)
        <class 'sage.interfaces.regina.ReginaElement'>
        sage: P = re.parent(); P
        Regina
        sage: type(P)
        <class 'sage.interfaces.regina.Regina'>

    Access to the Regina expression objects::

        sage: res = re._inst
        sage: type(res)
         <class 'regina.engine.GroupExpression'>

    Applying Regina methods::

        sage: re.cycleLeft(); re
        <regina.GroupExpression: g0^-1 g0 g1>

    Conversion to Sage::

        sage: re.sage() == b
        False
        sage: re.cycleRight()
        sage: re.sage() == b
        True

    TESTS::

        sage: p = regina.Polynomial([-2, ~7])
        sage: TestSuite(p).run(skip='_test_category')
    """

    _sage_parent = None  # for interface elements create from sage objects

    def _tab_completion(self):
        r"""
        Return a list of all methods of this object.

        EXAMPLES::

            sage: 'detail' in regina.AbelianGroup()._tab_completion()
            True
        """
        return dir(self._inst)

    def __getitem__(self, n):
        r"""
        EXAMPLES::

            sage: t = regina([-2, 7/3])
            sage: type(t)
            <class 'sage.interfaces.regina.ReginaElement'>
            sage: t[1]     # indirect doctest
            7/3
            sage: type(t[1])
            <class 'sage.interfaces.regina.ReginaElement'>
            sage: type(t[1]._inst)
            <class 'regina.engine.Rational'>
        """
        return self.parent().new(self._inst[n])

    def __iter__(self):
        r"""
        EXAMPLES::

            sage: list(regina([-2, 7/3]))
            [-2, 7/3]
        """
        for i in range(len(self)):
            yield self[i]

    def __getattr__(self, attrname):
        r"""
        EXAMPLES::

            sage: type(regina.AbelianGroup().detail)
            <class 'sage.interfaces.regina.ReginaFunctionElement'>
            sage: regina.AbelianGroup().detail._name
            <bound method pybind11_detail_function_record_v1_system_libstdcpp_gxx_abi_1xxx_use_cxx11_abi_1.detail of <regina.AbelianGroup: 0>>
        """
        P = self._check_valid()
        if attrname == '_inst':
            self._inst = P.get(self.name())
            return self._inst
        if attrname[:1] == "_":
            raise AttributeError
        else:
            inst = self._inst
            if hasattr(inst, attrname):
                attr = inst.__getattribute__(attrname)
                if callable(attr):
                    return ReginaFunctionElement(self, attr)
        raise AttributeError

    def __bool__(self):
        """
        Return whether this element is not ``False``.

        EXAMPLES::

            sage: M = regina.Matrix2(); M
            <regina.Matrix2: [[ 0 0 ] [ 0 0 ]]>
            sage: M.isIdentity()
            False
            sage: M.isZero()
            True
        """
        return bool(self._inst)

    def __deepcopy__(self, memo=None):
        r"""
        EXAMPLES::

            sage: C = CubicBraidGroup(3)
            sage: c = C.an_element()
            sage: cr = regina(c); cr
            <regina.GroupExpression: g0 g1>
            sage: from copy import deepcopy, copy
            sage: crd = deepcopy(cr)
            sage: crc = copy(cr)
            sage: cr.cycleRight(); cr
            <regina.GroupExpression: g1 g0>
            sage: (crd, crd == cr)
            (<regina.GroupExpression: g0 g1>, False)
            sage: (crc, crc == cr)
            (<regina.GroupExpression: g1 g0>, True)
        """
        P = self._check_valid()
        inst = self._inst
        new = inst.__class__(inst)
        res = P(new)
        res._sage_parent = self._sage_parent
        return res

    def _latex_(self):
        r"""
        EXAMPLES::

            sage: a = regina(~5)
            sage: latex(a)
            \frac{1}{5}
            sage: b = regina(BraidGroup(4)((1, 2, 3, -2, -1))); b
            <regina.GroupExpression: g0 g1 g2 g1^-1 g0^-1>
            sage: latex(b)
            g_{0}g_{1}g_{2}g_{1}^{-1}g_{0}^{-1}
        """
        if hasattr(self, 'tex'):
            return self.tex()
        return super()._latex_()

    def _repr_(self):
        r"""
        EXAMPLES::

            sage: Q = regina('GroupPresentation()')
            sage: repr(Q)
            '<regina.GroupPresentation: < >>'
        """
        return repr(self._inst)

    def _richcmp_(self, other, op):
        r"""
        Comparison of interface elements.

        EXAMPLES::

            sage: a = regina(3)
            sage: b = regina(5)
            sage: a < b
            True
            sage: a > b
            False
            sage: F3 = FreeGroup(3)
            sage: a, b, c = F3.gens()
            sage: f = a**2*b*~c
            sage: fr = regina(f)
            sage: cr = regina(c)
            sage: fr == cr
            False
            sage: fr != fr * cr
            True
        """
        from sage.structure.richcmp import rich_to_bool, op_EQ, op_NE
        if self._inst == other._inst:
            return rich_to_bool(op, 0)
        if op == op_EQ:
            return False
        if op == op_NE:
            return True
        try:
            if self._inst < other._inst:
                return rich_to_bool(op, -1)
            if self._inst > other._inst:
                return rich_to_bool(op, 1)
        except TypeError:
            pass
        return super()._richcmp_(other, op)

    def _operation(self, operation, other=None):
        r"""
        Return the result of applying the binary operation
        ``operation`` on the arguments ``self`` and ``other``, or the
        unary operation on ``self`` if ``other`` is not given.

        This is a utility function which factors out much of the
        commonality used in the arithmetic operations for interface
        elements.

        INPUT:

        - ``operation`` -- string representing the operation
          being performed; for example, '*', or '1/'

        - ``other`` -- the other operand; if ``other`` is ``None``,
          then the operation is assumed to be unary rather than binary

        OUTPUT: an interface element

        EXAMPLES::

            sage: l = regina(range(2, 5))
            sage: [type(i) for i in l]
            [<class 'sage.interfaces.regina.ReginaElement'>,
             <class 'sage.interfaces.regina.ReginaElement'>,
             <class 'sage.interfaces.regina.ReginaElement'>]
            sage: s = sum(l); s, type(s)
            (9, <class 'sage.interfaces.regina.ReginaElement'>)
            sage: p = prod(l); p, type(p)
            (24, <class 'sage.interfaces.regina.ReginaElement'>)
            sage: c = CubicBraidGroup(3).an_element()
            sage: cr = regina(c); cr
            <regina.GroupExpression: g0 g1>
            sage: cr**(-3)
            <regina.GroupExpression: g1^-1 g0^-1 g1^-1 g0^-1 g1^-1 g0^-1>
        """
        P = self._check_valid()
        sinst = self._inst
        oinst = other if other is None else other._inst

        def is_native(inst):
            return type(inst) in (int, float, complex)
        if operation in ('+', '*'):
            if is_native(sinst) and is_native(oinst):
                if operation == '*':
                    return P(sinst * oinst)
                return P(sinst + oinst)
            if type(sinst) is type(oinst):
                if hasattr(self, 'addTermsLast'):
                    new = self.__deepcopy__()
                    new.addTermsLast(other)
                    new.simplify()
                    return new
        if operation == '^':
            if is_native(sinst) and is_native(oinst):
                return P(sinst**oinst)
            try:
                exp = int(other)
            except TypeError:
                raise TypeError('only integer exponents allowed!')

            if exp == 1:
                return self
            if exp == 2:
                return self * self
            if exp > 0:
                for i in range(exp):
                    return self**(exp - 1) * self
            else:
                return (~self)**(-exp)
        if operation == '1/':
            if is_native(sinst):
                return P(1 / sinst)
            if hasattr(self, 'inverse'):
                return self.inverse()
        return super()._operation(operation, other=other)

    def _sage_(self, locals={}):
        r"""
        Attempt to return a Sage version of this object.

        This method works successfully when Regina returns a result
        or list of results that consist only of:

        - numbers, i.e. integers, floats, complex numbers;
        - functions and named constants also present in Sage, where:

          * Sage knows how to translate the function or constant's name
            from Regina's naming scheme, or
          * you provide a translation dictionary `locals`, or
          * the Sage name for the function or constant is simply the
            Regina name in lower case;

        - symbolic variables whose names do not pathologically overlap with
          objects already defined in Sage.

        This method will not work when Regina's output includes:

        - strings;
        - functions unknown to Sage;
        - Regina functions with different parameters/parameter order to
          the Sage equivalent. In this case, define a function to do the
          parameter conversion, and pass it in via the locals dictionary.

        EXAMPLES::

            sage: p = regina("Laurent2([(1,2,3), (2,-1,5)])"); p
            <regina.Laurent2: 5 x^2 y^-1 + 3 x y^2>
            sage: p.sage()
            3*x*y^2 + 5*x^2*y^-1
            sage: R.<u> = PolynomialRing(ZZ)
            sage: p = u**3 -2*u + 9
            sage: rp = regina(p); rp
            <regina.PolynomialRational: x^3 - 2 x + 9>
            sage: rp.sage() == p
            True
            sage: F3 = FreeGroup(3)
            sage: a, b, c = F3.gens()
            sage: f = a**2*b*~c
            sage: fr = regina(f); fr
            <regina.GroupExpression: g0^2 g1 g2^-1>
            sage: fr.sage() == f
            True
        """
        def from_detail_str(lc):
            r"""
            Regina provides a detail method for many of its classes.
            Here we try to use it to convert back to Sage.
            """
            if locals:
                lc.update(locals)
            from sage.misc.sage_eval import sage_eval
            from sage.repl.preparse import implicit_mul
            s = self.detail().split('\n')[0]
            s = s.replace(' ', '')
            v = list(lc)
            for i in v + ['(']:
                for j in v + [')']:
                    s = s.replace('%s%s' % (i, j), '%s*%s' % (i, j))
            s = implicit_mul(s)
            return sage_eval(s, locals=lc)

        P = self._check_valid()
        inst = self._inst
        nspc = P._namespace
        if hasattr(inst, 'sage'):
            return inst.sage()
        if isinstance(inst, (nspc.Polynomial, nspc.Laurent, nspc.Laurent2)):
            if self._sage_parent:
                R = self._sage_parent
                old_var_names = ['x', 'y']
                new_var_names = R.variable_names()
                lc = {old_var_names[i]: R.gens_dict()[new_var_names[i]] for i in range(len(new_var_names))}
            elif isinstance(inst, nspc.Polynomial):
                from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
                from sage.rings.integer_ring import ZZ
                R = PolynomialRing(ZZ, 'x')
                lc = R.gens_dict()
            else:
                from sage.rings.polynomial.laurent_polynomial_ring import LaurentPolynomialRing
                from sage.rings.integer_ring import ZZ
                if isinstance(inst, nspc.Laurent):
                    R = LaurentPolynomialRing(ZZ, 'x')
                else:
                    R = LaurentPolynomialRing(ZZ, 'x, y')
                lc = R.gens_dict()
            return from_detail_str(lc)
        if isinstance(inst, nspc.GroupExpression):
            num_gens = max(t.generator for t in inst.terms()) + 1
            if self._sage_parent:
                F = self._sage_parent
            else:
                from sage.groups.free_group import FreeGroup
                F = FreeGroup(num_gens)
            gens = F.gens()
            lc = {'g%s' % i: gens[i] for i in range(num_gens)}
            return from_detail_str(lc)
        if isinstance(inst, nspc.Link):
            from sage.knots.link import Link
            return Link(inst.pdData())
        if hasattr(self, 'detail'):
            return from_detail_str(locals)
        if locals:
            # if locals are given we use `_sage_repr`
            # surely this only covers simple cases
            from sage.misc.sage_eval import sage_eval
            return sage_eval(self._sage_repr(), locals=locals)
        return inst


@instancedoc
class ReginaFunctionElement(InterfaceFunctionElement):
    r"""
    Interface methods of interface elements.

    EXAMPLES::

        sage: A = regina.AbelianGroup()
        sage: A.addRank
        <bound method pybind11_detail_function_record_v1_system_libstdcpp_gxx_abi_1xxx_use_cxx11_abi_1.addRank of <regina.AbelianGroup: 0>>
        sage: type(A.addRank)
        <class 'sage.interfaces.regina.ReginaFunctionElement'>
    """
    def __call__(self, *args, **kwds):
        r"""
        Call this function with the given args and kwds.

        This method is overloaded since the functions are
        Python functions which don't need to be interpreted
        from strings.

        EXAMPLES::

            sage: A = regina.AbelianGroup(); A
            <regina.AbelianGroup: 0>
            sage: a = A.addRank
            sage: a(); A
            <regina.AbelianGroup: Z>
            sage: a(int(2)); A
            <regina.AbelianGroup: 3 Z>
        """
        P = self._obj.parent()
        return P._function_call(self._name, *args, **kwds)


@instancedoc
class ReginaFunction(InterfaceFunction):
    r"""
    Interface Function.

    EXAMPLES::

        sage: m = regina.MatrixInt; m
        <class 'regina.engine.MatrixInt'>
        sage: type(m)
        <class 'sage.interfaces.regina.ReginaFunction'>
    """
    def __call__(self, *args, **kwds):
        r"""
        Call this function with the given args and kwds.

        This method is overloaded since the functions are
        Python functions which don't need to be interpreted
        from strings.

        EXAMPLES::

            sage: m = regina.MatrixInt
            sage: m([[1, 2], [3, 4]])
            <regina.MatrixInt: [[ 1 2 ] [ 3 4 ]]>
        """
        P = self._parent
        return P._function_call(self._name, *args, **kwds)


# An instance
regina = Regina()



# distutils: depends = NTL/ZZ.h
# distutils: libraries = NTL_LIBRARIES gmp
# distutils: extra_compile_args = NTL_CFLAGS
# distutils: include_dirs = NTL_INCDIR
# distutils: library_dirs = NTL_LIBDIR
# distutils: extra_link_args = NTL_LIBEXTRA
# distutils: language = c++
"""
Conversion between NTL's ``ZZ`` and various other types
"""

#*****************************************************************************
#       Copyright (C) 2015 Jeroen Demeyer <jdemeyer@cage.ugent.be>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.libs.gmp.mpz cimport mpz_init, mpz_clear
from sage.libs.gmp.pylong cimport mpz_set_pylong

cdef extern from "ntlwrap_impl.h":
    void ZZ_to_mpz(mpz_t output, ZZ_c* x)
    void mpz_to_ZZ(ZZ_c *output, mpz_srcptr x)

cdef void PyLong_to_ZZ(ZZ_c* z, value) noexcept:
    """
    Convert ``value`` (which must be a Python ``long``) to NTL.
    """
    cdef mpz_t t
    mpz_init(t)
    mpz_set_pylong(t, value)
    mpz_to_ZZ(z, t)
    mpz_clear(t)



r"""
Design notes

The implementation depends on the fact that we have dimension formulas
(see ``dims.py``) for spaces of modular forms with character, and new
subspaces, so that we don't have to compute `q`-expansions for the whole
space in order to compute `q`-expansions / elements / and dimensions of
certain subspaces.  Also, the following design is much simpler than
the one I used in MAGMA because submodules don't have lots of
complicated special labels.  A modular forms module can consist of the
span of any elements; they need not be Hecke equivariant or anything
else.

The internal basis of `q`-expansions of modular forms for the ambient
space is defined as follows::

  First Block:   Cuspidal Subspace
  Second Block:  Eisenstein Subspace

      Cuspidal Subspace:    Block for each level `M` dividing `N`, from highest
                            level to lowest. The block for level `M`
                            contains the images at level `N` of the
                            newsubspace of level `M` (basis, then
                            basis(q**d), then basis(q**e), etc.)

     Eisenstein Subspace:   characters, etc.

Since we can compute dimensions of cuspidal subspaces quickly and
easily, it should be easy to locate any of the above blocks.  Hence,
e.g., to compute basis for new cuspidal subspace, just have to return
first n standard basis vector where n is the dimension.  However, we
can also create completely arbitrary subspaces as well.

The base ring is the ring generated by the character values (or
bigger).  In MAGMA the base was always `\ZZ`, which is confusing.
"""

#########################################################################
#       Copyright (C) 2006 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  http://www.gnu.org/licenses/
#########################################################################



"""
Homology Groups

This module defines a :meth:`HomologyGroup` class which is an abelian
group that prints itself in a way that is suitable for homology
groups.
"""
########################################################################
#       Copyright (C) 2013 John H. Palmieri <palmieri@math.washington.edu>
#                          Volker Braun <vbraun.name@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#
#                  https://www.gnu.org/licenses/
########################################################################

from sage.modules.free_module import VectorSpace
from sage.groups.additive_abelian.additive_abelian_group import AdditiveAbelianGroup_fixed_gens
from sage.rings.integer_ring import ZZ


class HomologyGroup_class(AdditiveAbelianGroup_fixed_gens):
    """
    Discrete Abelian group on `n` generators. This class inherits from
    :class:`~sage.groups.additive_abelian.additive_abelian_group.AdditiveAbelianGroup_fixed_gens`;
    see :mod:`sage.groups.additive_abelian.additive_abelian_group` for more
    documentation. The main difference between the classes is in the print
    representation.

    EXAMPLES::

        sage: from sage.homology.homology_group import HomologyGroup
        sage: G = AbelianGroup(5, [5,5,7,8,9]); G                                       # needs sage.groups
        Multiplicative Abelian group isomorphic to C5 x C5 x C7 x C8 x C9
        sage: H = HomologyGroup(5, ZZ, [5,5,7,8,9]); H
        C5 x C5 x C7 x C8 x C9
        sage: G == loads(dumps(G))                                                      # needs sage.groups
        True
        sage: AbelianGroup(4)                                                           # needs sage.groups
        Multiplicative Abelian group isomorphic to Z x Z x Z x Z
        sage: HomologyGroup(4, ZZ)
        Z x Z x Z x Z
        sage: HomologyGroup(100, ZZ)
        Z^100
    """
    def __init__(self, n, invfac) -> None:
        """
        See :func:`HomologyGroup` for full documentation.

        EXAMPLES::

            sage: from sage.homology.homology_group import HomologyGroup
            sage: H = HomologyGroup(5, ZZ, [5,5,7,8,9]); H
            C5 x C5 x C7 x C8 x C9
        """
        n = len(invfac)
        A = ZZ ** n
        B = A.span([A.gen(i) * invfac[i] for i in range(n)])

        AdditiveAbelianGroup_fixed_gens.__init__(self, A, B, A.gens())
        self._original_invts = invfac

    def _repr_(self) -> str:
        """
        Print representation of ``self``.

        EXAMPLES::

            sage: from sage.homology.homology_group import HomologyGroup
            sage: H = HomologyGroup(7, ZZ, [4,4,4,4,4,7,7])
            sage: H._repr_()
            'C4^5 x C7 x C7'
            sage: HomologyGroup(6, ZZ)
            Z^6
        """
        eldv = self._original_invts
        if len(eldv) == 0:
            return "0"
        rank = len([x for x in eldv if x == 0])
        torsion = sorted(x for x in eldv if x)
        if rank > 4:
            g = ["Z^%s" % rank]
        else:
            g = ["Z"] * rank
        if len(torsion) != 0:
            printed = []
            for t in torsion:
                numfac = torsion.count(t)
                too_many = (numfac > 4)
                if too_many:
                    if t not in printed:
                        g.append("C{}^{}".format(t, numfac))
                        printed.append(t)
                else:
                    g.append("C%s" % t)
        times = " x "
        return times.join(g)

    def _latex_(self):
        r"""
        LaTeX representation of ``self``.

        EXAMPLES::

            sage: from sage.homology.homology_group import HomologyGroup
            sage: H = HomologyGroup(7, ZZ, [4,4,4,4,4,7,7])
            sage: H._latex_()
            'C_{4}^{5} \\times C_{7} \\times C_{7}'
            sage: latex(HomologyGroup(6, ZZ))
            \ZZ^{6}
        """
        eldv = self._original_invts
        if len(eldv) == 0:
            return "0"
        rank = len([x for x in eldv if x == 0])
        torsion = sorted(x for x in eldv if x)
        if rank > 4:
            g = ["\\ZZ^{{{}}}".format(rank)]
        else:
            g = ["\\ZZ"] * rank
        if torsion:
            printed = []
            for t in torsion:
                numfac = torsion.count(t)
                too_many = (numfac > 4)
                if too_many:
                    if t not in printed:
                        g.append("C_{{{}}}^{{{}}}".format(t, numfac))
                        printed.append(t)
                else:
                    g.append("C_{{{}}}".format(t))
        times = " \\times "
        return times.join(g)


def HomologyGroup(n, base_ring, invfac=None):
    """
    Abelian group on `n` generators which represents a homology group in a
    fixed degree.

    INPUT:

    - ``n`` -- integer; the number of generators

    - ``base_ring`` -- ring; the base ring over which the homology is computed

    - ``inv_fac`` -- list of integers; the invariant factors -- ignored
      if the base ring is a field

    OUTPUT:

    A class that can represent the homology group in a fixed
    homological degree.

    EXAMPLES::

        sage: from sage.homology.homology_group import HomologyGroup
        sage: G = AbelianGroup(5, [5,5,7,8,9]); G                                       # needs sage.groups
        Multiplicative Abelian group isomorphic to C5 x C5 x C7 x C8 x C9
        sage: H = HomologyGroup(5, ZZ, [5,5,7,8,9]); H
        C5 x C5 x C7 x C8 x C9
        sage: AbelianGroup(4)                                                           # needs sage.groups
        Multiplicative Abelian group isomorphic to Z x Z x Z x Z
        sage: HomologyGroup(4, ZZ)
        Z x Z x Z x Z

        sage: # needs sage.libs.flint (otherwise timeout)
        sage: HomologyGroup(100, ZZ)
        Z^100
    """
    if base_ring.is_field():
        return VectorSpace(base_ring, n)

    # copied from AbelianGroup:
    if invfac is None:
        if isinstance(n, (list, tuple)):
            invfac = n
            n = len(n)
        else:
            invfac = []
    if len(invfac) < n:
        invfac = [0] * (n - len(invfac)) + invfac
    elif len(invfac) > n:
        raise ValueError("invfac (={}) must have length n (={})".format(invfac, n))
    return HomologyGroup_class(n, invfac)




from sage.plot.plot3d.plot3d import plot3d, cylindrical_plot3d, spherical_plot3d, Spherical, SphericalElevation, Cylindrical
from sage.plot.plot3d.parametric_plot3d import parametric_plot3d
from sage.plot.plot3d.plot_field3d import plot_vector_field3d

# We lazy_import the following modules since they import numpy which slows down sage startup
from sage.misc.lazy_import import lazy_import
lazy_import("sage.plot.plot3d.implicit_plot3d", ["implicit_plot3d"])

from sage.plot.plot3d.list_plot3d import list_plot3d
from sage.plot.plot3d.revolution_plot3d import revolution_plot3d

from sage.plot.plot3d.platonic import tetrahedron, cube, octahedron, dodecahedron, icosahedron

from sage.plot.plot3d.shapes2 import sphere, line3d, polygon3d, polygons3d, point3d, text3d, bezier3d

from sage.plot.plot3d.shapes import arrow3d

# from shapes import Box, ColorCube, Cone, Cylinder, LineSegment, Arrow, Sphere, Torus, Text as Text3D
# from parametric_surface import ParametricSurface, MoebiusStrip
# from plot3d import plot3d, axes as axes3d
del lazy_import



# sage.doctest: optional - numpy
r"""
Dense integer vectors using a NumPy backend.

EXAMPLES::

    sage: from sage.modules.vector_numpy_integer_dense import Vector_numpy_integer_dense
    sage: v = Vector_numpy_integer_dense(FreeModule(ZZ, 3), [0, 0, 0]); v
    (0, 0, 0)
    sage: v[1] = 42
    sage: v
    (0, 42, 0)
    sage: v.numpy(dtype="int8")
    array([ 0, 42,  0], dtype=int8)

TESTS::

    sage: Vector_numpy_integer_dense(FreeModule(GF(2), 3), [0, 0, 0])
    Traceback (most recent call last):
    ...
    ValueError: parent must have base ring ZZ
"""

# ****************************************************************************
#       Copyright (C) 2021      Matthias Koeppe
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

cimport numpy
import numpy

from sage.rings.integer_ring import ZZ

# This is for the NumPy C API (the PyArray... functions) to work
numpy.import_array()


cdef class Vector_numpy_integer_dense(Vector_numpy_dense):

    def __cinit__(self, parent, entries, coerce=True, copy=True):
        if parent.base_ring() is not ZZ:
            raise ValueError("parent must have base ring ZZ")
        self._numpy_dtype = numpy.dtype('int64')
        self._numpy_dtypeint = numpy.NPY_INT64
        self._python_dtype = int
        self._sage_dtype = ZZ
        self.__create_vector__()
        return



# sage.doctest: needs sage.libs.flint sage.libs.pari
"""
Torsion subgroups of modular abelian varieties

Sage can compute information about the structure of the torsion
subgroup of a modular abelian variety. Sage computes a multiple of
the order by computing the greatest common divisor of the orders of
the torsion subgroup of the reduction of the abelian variety modulo
p for various primes p. Sage computes a divisor of the order by
computing the rational cuspidal subgroup. When these two bounds
agree (which is often the case), we determine the exact structure
of the torsion subgroup.

AUTHORS:

- William Stein (2007-03)

EXAMPLES: First we consider `J_0(50)` where everything
works out nicely::

    sage: J = J0(50)
    sage: T = J.rational_torsion_subgroup(); T
    Torsion subgroup of Abelian variety J0(50) of dimension 2
    sage: T.multiple_of_order()
    15
    sage: T.divisor_of_order()
    15
    sage: T.gens()
    ([(1/15, 3/5, 2/5, 14/15)],)
    sage: T.invariants()
    [15]
    sage: d = J.decomposition(); d
    [Simple abelian subvariety 50a(1,50) of dimension 1 of J0(50),
     Simple abelian subvariety 50b(1,50) of dimension 1 of J0(50)]
    sage: d[0].rational_torsion_subgroup().order()
    3
    sage: d[1].rational_torsion_subgroup().order()
    5

Next we make a table of the upper and lower bounds for each new
factor.

::

    sage: for N in range(1,38):
    ....:    for A in J0(N).new_subvariety().decomposition():
    ....:        T = A.rational_torsion_subgroup()
    ....:        print('%-5s%-5s%-5s%-5s'%(N, A.dimension(), T.divisor_of_order(), T.multiple_of_order()))
    11   1    5    5
    14   1    6    6
    15   1    8    8
    17   1    4    4
    19   1    3    3
    20   1    6    6
    21   1    8    8
    23   2    11   11
    24   1    8    8
    26   1    3    3
    26   1    7    7
    27   1    3    3
    29   2    7    7
    30   1    6    6
    31   2    5    5
    32   1    4    4
    33   1    4    4
    34   1    6    6
    35   1    3    3
    35   2    16   16
    36   1    6    6
    37   1    1    1
    37   1    3    3

TESTS::

    sage: T = J0(54).rational_torsion_subgroup()
    sage: loads(dumps(T)) == T
    True
"""

# ****************************************************************************
#       Copyright (C) 2007 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.arith.misc import divisors, gcd
from sage.misc.misc_c import prod
from sage.modular.abvar.torsion_point import TorsionPoint
from sage.modular.arithgroup.congroup_gamma0 import Gamma0_class
from sage.modular.arithgroup.congroup_gamma1 import Gamma1_class
from sage.modular.dirichlet import DirichletGroup
from sage.modules.module import Module
from sage.rings.fast_arith import prime_range
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.sets.primes import Primes
from sage.structure.richcmp import richcmp_method, richcmp

from .finite_subgroup import FiniteSubgroup


@richcmp_method
class RationalTorsionSubgroup(FiniteSubgroup):
    """
    The torsion subgroup of a modular abelian variety.
    """
    def __init__(self, abvar):
        """
        Create the torsion subgroup.

        INPUT:

        - ``abvar`` -- a modular abelian variety

        EXAMPLES::

            sage: T = J0(14).rational_torsion_subgroup(); T
            Torsion subgroup of Abelian variety J0(14) of dimension 1
            sage: type(T)
            <class 'sage.modular.abvar.torsion_subgroup.RationalTorsionSubgroup_with_category'>
        """
        FiniteSubgroup.__init__(self, abvar)

    def _repr_(self):
        """
        Return string representation of this torsion subgroup.

        EXAMPLES::

            sage: T = J1(13).rational_torsion_subgroup(); T
            Torsion subgroup of Abelian variety J1(13) of dimension 2
            sage: T._repr_()
            'Torsion subgroup of Abelian variety J1(13) of dimension 2'
        """
        return "Torsion subgroup of %s" % self.abelian_variety()

    def __richcmp__(self, other, op):
        """
        Compare torsion subgroups.

        INPUT:

        - ``other`` -- an object

        If other is a torsion subgroup, the abelian varieties are compared.
        Otherwise, the generic behavior for finite abelian variety
        subgroups is used.

        EXAMPLES::

            sage: G = J0(11).rational_torsion_subgroup(); H = J0(13).rational_torsion_subgroup()
            sage: G == G
            True
            sage: G < H   # since 11 < 13
            True
            sage: G > H
            False
        """
        if isinstance(other, RationalTorsionSubgroup):
            return richcmp(self.abelian_variety(), other.abelian_variety(), op)
        return FiniteSubgroup.__richcmp__(self, other, op)

    def order(self, proof=True):
        """
        Return the order of the torsion subgroup of this modular abelian
        variety.

        This function may fail if the multiple obtained by counting points
        modulo `p` exceeds the divisor obtained from the rational cuspidal
        subgroup.

        The computation of the rational torsion order of J1(p) is conjectural
        and will only be used if ``proof=False``. See Section 6.2.3 of [CES2003]_.

        INPUT:

        - ``proof`` -- boolean (default: ``True``)

        OUTPUT: the order of this torsion subgroup

        EXAMPLES::

            sage: A = J0(11)
            sage: A.rational_torsion_subgroup().order()
            5
            sage: A = J0(23)
            sage: A.rational_torsion_subgroup().order()
            11
            sage: T = J0(37)[1].rational_torsion_subgroup()
            sage: T.order()
            3

            sage: J = J1(13)
            sage: J.rational_torsion_subgroup().order()
            19

        Sometimes the order can only be computed with ``proof=False``. ::

            sage: J = J1(23)
            sage: J.rational_torsion_subgroup().order()
            Traceback (most recent call last):
            ...
            RuntimeError: Unable to compute order of torsion subgroup
            (it is in [408991, 9406793])

            sage: J.rational_torsion_subgroup().order(proof=False)
            408991
        """
        O = self.possible_orders(proof=proof)
        if len(O) == 1:
            n = O[0]
            self._order = n
            return n
        raise RuntimeError("Unable to compute order of torsion subgroup (it is in %s)" % O)

    def lattice(self):
        """
        Return lattice that defines this torsion subgroup, if possible.

        .. warning::

           There is no known algorithm in general to compute the
           rational torsion subgroup. Use rational_cusp_group to
           obtain a subgroup of the rational torsion subgroup in
           general.

        EXAMPLES::

            sage: J0(11).rational_torsion_subgroup().lattice()
            Free module of degree 2 and rank 2 over Integer Ring
            Echelon basis matrix:
            [  1   0]
            [  0 1/5]

        The following fails because in fact I know of no (reasonable)
        algorithm to provably compute the torsion subgroup in general.

        ::

            sage: T = J0(33).rational_torsion_subgroup()
            sage: T.lattice()
            Traceback (most recent call last):
            ...
            NotImplementedError: unable to compute the rational torsion subgroup
            in this case (there is no known general algorithm yet)

        The problem is that the multiple of the order obtained by counting
        points over finite fields is twice the divisor of the order got
        from the rational cuspidal subgroup.

        ::

            sage: T.multiple_of_order(30)
            200
            sage: J0(33).rational_cusp_subgroup().order()
            100
        """
        A = self.abelian_variety()
        if A.dimension() == 0:
            return []
        R = A.rational_cusp_subgroup()
        if R.order() == self.multiple_of_order():
            return R.lattice()
        raise NotImplementedError("unable to compute the rational torsion subgroup in this case (there is no known general algorithm yet)")

    def possible_orders(self, proof=True):
        """
        Return the possible orders of this torsion subgroup. Outside of special
        cases, this is done by computing a divisor and multiple of the order.

        INPUT:

        - ``proof`` -- boolean (default: ``True``)

        OUTPUT: an array of positive integers

        The computation of the rational torsion order of J1(p) is conjectural
        and will only be used if ``proof=False``. See Section 6.2.3 of [CES2003]_.

        EXAMPLES::

            sage: J0(11).rational_torsion_subgroup().possible_orders()
            [5]
            sage: J0(33).rational_torsion_subgroup().possible_orders()
            [100, 200]

            sage: J1(13).rational_torsion_subgroup().possible_orders()
            [19]
            sage: J1(16).rational_torsion_subgroup().possible_orders()
            [1, 2, 4, 5, 10, 20]
        """
        try:
            if proof:
                return self._possible_orders
            return self._possible_orders_proof_false
        except AttributeError:
            pass

        A = self.abelian_variety()
        N = A.level()
        # return the order of the cuspidal subgroup in the J0(p) case
        if A.is_J0() and N.is_prime():
            self._possible_orders = [QQ((A.level()-1)/12).numerator()]
            self._possible_orders_proof_false = self._possible_orders
            return self._possible_orders

        # the elliptic curve case
        if A.dimension() == 1:
            self._possible_orders = [A.elliptic_curve().torsion_order()]
            self._possible_orders_proof_false = self._possible_orders
            return self._possible_orders

        # the conjectural J1(p) case
        if not proof and A.is_J1() and N.is_prime():
            epsilons = [epsilon for epsilon in DirichletGroup(N)
                        if not epsilon.is_trivial() and epsilon.is_even()]
            bernoullis = [epsilon.bernoulli(2) for epsilon in epsilons]
            self._possible_orders_proof_false = [ZZ(N/(2**(N-3))*prod(bernoullis))]
            return self._possible_orders_proof_false

        u = self.multiple_of_order()
        l = self.divisor_of_order()

        assert u % l == 0
        O = [l * d for d in divisors(u//l)]
        self._possible_orders = O
        if u == l:
            self._possible_orders_proof_false = O
        return O

    def divisor_of_order(self):
        """
        Return a divisor of the order of this torsion subgroup of a modular
        abelian variety.

        OUTPUT: a divisor of this torsion subgroup

        EXAMPLES::

            sage: t = J0(37)[1].rational_torsion_subgroup()
            sage: t.divisor_of_order()
            3

            sage: J = J1(19)
            sage: J.rational_torsion_subgroup().divisor_of_order()
            4383

            sage: J = J0(45)
            sage: J.rational_cusp_subgroup().order()
            32
            sage: J.rational_cuspidal_subgroup().order()
            64
            sage: J.rational_torsion_subgroup().divisor_of_order()
            64
        """
        try:
            return self._divisor_of_order
        except AttributeError:
            pass

        A = self.abelian_variety()
        N = A.level()

        if A.dimension() == 0:
            self._divisor_of_order = ZZ(1)
            return self._divisor_of_order

        # return the order of the cuspidal subgroup in the J0(p) case
        if A.is_J0() and N.is_prime():
            self._divisor_of_order = QQ((A.level()-1)/12).numerator()
            return self._divisor_of_order

        # The elliptic curve case
        if A.dimension() == 1:
            self._divisor_of_order = A.elliptic_curve().torsion_order()
            return self._divisor_of_order

        # The J1(p) case
        if A.is_J1() and N.is_prime():
            epsilons = [epsilon for epsilon in DirichletGroup(N)
                        if not epsilon.is_trivial() and epsilon.is_even()]
            bernoullis = [epsilon.bernoulli(2) for epsilon in epsilons]
            self._divisor_of_order = ZZ(N/(2**(N-3))*prod(bernoullis))
            return self._divisor_of_order

        # The Gamma0 case
        if all(isinstance(G, Gamma0_class) for G in A.groups()):
            self._divisor_of_order = A.rational_cuspidal_subgroup().order()
            return self._divisor_of_order

        # Unhandled case
        self._divisor_of_order = ZZ(1)
        return self._divisor_of_order

    def multiple_of_order(self, maxp=None, proof=True):
        """
        Return a multiple of the order.

        INPUT:

        - ``proof`` -- boolean (default: ``True``)

        The computation of the rational torsion order of J1(p) is conjectural
        and will only be used if proof=False. See Section 6.2.3 of [CES2003]_.

        EXAMPLES::

            sage: J = J1(11); J
            Abelian variety J1(11) of dimension 1
            sage: J.rational_torsion_subgroup().multiple_of_order()
            5

            sage: J = J0(17)
            sage: J.rational_torsion_subgroup().order()
            4

        This is an example where proof=False leads to a better bound and better
        performance. ::

            sage: J = J1(23)
            sage: J.rational_torsion_subgroup().multiple_of_order()  # long time (2s)
            9406793
            sage: J.rational_torsion_subgroup().multiple_of_order(proof=False)
            408991
        """

        try:
            if proof:
                return self._multiple_of_order
            return self._multiple_of_order_proof_false
        except AttributeError:
            pass

        A = self.abelian_variety()
        N = A.level()

        if A.dimension() == 0:
            self._multiple_of_order = ZZ(1)
            self._multiple_of_order_proof_false = self._multiple_of_order
            return self._multiple_of_order

        # return the order of the cuspidal subgroup in the J0(p) case
        if A.is_J0() and N.is_prime():
            self._multiple_of_order = QQ((A.level()-1)/12).numerator()
            self._multiple_of_order_proof_false = self._multiple_of_order
            return self._multiple_of_order

        # The elliptic curve case
        if A.dimension() == 1:
            self._multiple_of_order = A.elliptic_curve().torsion_order()
            self._multiple_of_order_proof_false = self._multiple_of_order
            return self._multiple_of_order

        # The conjectural J1(p) case
        if not proof and A.is_J1() and N.is_prime():
            epsilons = [epsilon for epsilon in DirichletGroup(N)
                        if not epsilon.is_trivial() and epsilon.is_even()]
            bernoullis = [epsilon.bernoulli(2) for epsilon in epsilons]
            self._multiple_of_order_proof_false = ZZ(N/(2**(N-3))*prod(bernoullis))
            return self._multiple_of_order_proof_false

        # The Gamma0 and Gamma1 case
        if all(isinstance(G, (Gamma0_class, Gamma1_class)) for G in A.groups()):
            self._multiple_of_order = self.multiple_of_order_using_frobp()
            return self._multiple_of_order

        # Unhandled case
        raise NotImplementedError("No implemented algorithm")

    def multiple_of_order_using_frobp(self, maxp=None):
        """
        Return a multiple of the order of this torsion group.

        In the `Gamma_0` case, the multiple is computed using characteristic
        polynomials of Hecke operators of odd index not dividing the level. In
        the `Gamma_1` case, the multiple is computed by expressing the
        frobenius polynomial in terms of the characteristic polynomial of left
        multiplication by `a_p` for odd primes p not dividing the level.

        INPUT:

        - ``maxp`` -- (default: ``None``) if ``maxp`` is ``None``, return gcd
          of best bound computed so far with bound obtained by computing GCD's
          of orders modulo `p` until this gcd stabilizes for 3 successive
          primes. If ``maxp`` is given, just use all primes up to and including
          ``maxp``.

        EXAMPLES::

            sage: J = J0(11)
            sage: G = J.rational_torsion_subgroup()
            sage: G.multiple_of_order_using_frobp(11)
            5

        Increasing maxp may yield a tighter bound. If maxp=None, then Sage
        will use more primes until the multiple stabilizes for 3 successive
        primes.  ::

            sage: J = J0(389)
            sage: G = J.rational_torsion_subgroup(); G
            Torsion subgroup of Abelian variety J0(389) of dimension 32
            sage: G.multiple_of_order_using_frobp()
            97
            sage: [G.multiple_of_order_using_frobp(p) for p in prime_range(3,11)]
            [92645296242160800, 7275, 291]
            sage: [G.multiple_of_order_using_frobp(p) for p in prime_range(3,13)]
            [92645296242160800, 7275, 291, 97]
            sage: [G.multiple_of_order_using_frobp(p) for p in prime_range(3,19)]
            [92645296242160800, 7275, 291, 97, 97, 97]

        We can compute the multiple of order of the torsion subgroup for Gamma0
        and Gamma1 varieties, and their products. ::

            sage: A = J0(11) * J0(33)
            sage: A.rational_torsion_subgroup().multiple_of_order_using_frobp()
            1000

            sage: A = J1(23)
            sage: A.rational_torsion_subgroup().multiple_of_order_using_frobp()
            9406793
            sage: A.rational_torsion_subgroup().multiple_of_order_using_frobp(maxp=50)
            408991

            sage: A = J1(19) * J0(21)
            sage: A.rational_torsion_subgroup().multiple_of_order_using_frobp()
            35064

        The next example illustrates calling this function with a larger
        input and how the result may be cached when maxp is None::

            sage: T = J0(43)[1].rational_torsion_subgroup()
            sage: T.multiple_of_order_using_frobp()
            14
            sage: T.multiple_of_order_using_frobp(50)
            7
            sage: T.multiple_of_order_using_frobp()
            7

        This function is not implemented for general congruence subgroups
        unless the dimension is zero. ::

            sage: A = JH(13,[2]); A
            Abelian variety J0(13) of dimension 0
            sage: A.rational_torsion_subgroup().multiple_of_order_using_frobp()
            1

            sage: A = JH(15, [2]); A
            Abelian variety JH(15,[2]) of dimension 1
            sage: A.rational_torsion_subgroup().multiple_of_order_using_frobp()
            Traceback (most recent call last):
            ...
            NotImplementedError: torsion multiple only implemented for Gamma0 and Gamma1
        """
        if maxp is None:
            try:
                return self.__multiple_of_order_using_frobp
            except AttributeError:
                pass
        A = self.abelian_variety()
        if A.dimension() == 0:
            T = ZZ.one()
            self.__multiple_of_order_using_frobp = T
            return T
        if not all(isinstance(G, (Gamma0_class, Gamma1_class)) for G in A.groups()):
            raise NotImplementedError("torsion multiple only implemented for Gamma0 and Gamma1")

        bnd = ZZ.zero()
        N = A.level()
        cnt = 0
        if maxp is None:
            X = Primes()
        else:
            X = prime_range(maxp+1)
        for p in X:
            if (2*N) % p == 0:
                continue

            if (len(A.groups()) == 1 and isinstance(A.groups()[0], Gamma0_class)):
                f = A.hecke_polynomial(p)
                b = ZZ(f(p+1))
            else:
                from .constructor import AbelianVariety
                D = [AbelianVariety(f) for f in
                     A.newform_decomposition('a')]
                b = 1
                for simple in D:
                    G = simple.newform_level()[1]
                    if isinstance(G, Gamma0_class):
                        f = simple.hecke_polynomial(p)
                        b *= ZZ(f(p+1))
                    else:
                        f = simple.newform('a')
                        Kf = f.base_ring()
                        eps = f.character()
                        Qe = eps.base_ring()

                        if Kf != QQ:
                            # relativize number fields to compute charpoly of
                            # left multiplication of ap on Kf as a Qe-vector
                            # space.
                            Lf = Kf.relativize(Qe.gen(), 'a')
                            to_Lf = Lf.structure()[1]

                            name = Kf._names[0]
                            ap = to_Lf(f.modular_symbols(1).eigenvalue(p, name))

                            G_ps = ap.matrix().charpoly()
                            b *= ZZ(Qe(G_ps(1 + to_Lf(eps(p))*p)).norm())
                        else:
                            ap = f.modular_symbols(1).eigenvalue(p)
                            b *= ZZ(1 + eps(p)*p - ap)

            if bnd == 0:
                bnd = b
            else:
                bnd_last = bnd
                bnd = ZZ(gcd(bnd, b))
                if bnd == bnd_last:
                    cnt += 1
                else:
                    cnt = 0
                if maxp is None and cnt >= 2:
                    break

        # The code below caches the computed bound and
        # will be used if this function is called
        # again with maxp equal to None (the default).
        if maxp is None:
            # maxp is None but self.__multiple_of_order_using_frobp  is
            # not set, since otherwise we would have immediately
            # returned at the top of this function
            self.__multiple_of_order_using_frobp = bnd
        else:
            # maxp is given -- record new info we get as
            # a gcd...
            try:
                self.__multiple_of_order_using_frobp = \
                        gcd(self.__multiple_of_order_using_frobp, bnd)
            except AttributeError:
                # ... except in the case when
                # self.__multiple_of_order_using_frobp was never set.  In this
                # case, we just set it as long as the gcd stabilized for 3 in a
                # row.
                if cnt >= 2:
                    self.__multiple_of_order_using_frobp = bnd
        return bnd


class QQbarTorsionSubgroup(Module):

    Element = TorsionPoint

    def __init__(self, abvar):
        """
        Group of all torsion points over the algebraic closure on an
        abelian variety.

        INPUT:

        - ``abvar`` -- an abelian variety

        EXAMPLES::

            sage: A = J0(23)
            sage: A.qbar_torsion_subgroup()                                             # needs sage.rings.number_field
            Group of all torsion points in QQbar on Abelian variety J0(23) of dimension 2
        """
        self.__abvar = abvar
        Module.__init__(self, ZZ)

    def _repr_(self):
        """
        Print representation of QQbar points.

        OUTPUT: string

        EXAMPLES::

            sage: J0(23).qbar_torsion_subgroup()._repr_()                               # needs sage.rings.number_field
            'Group of all torsion points in QQbar on Abelian variety J0(23) of dimension 2'
        """
        return 'Group of all torsion points in QQbar on %s' % self.__abvar

    def field_of_definition(self):
        """
        Return the field of definition of this subgroup. Since this is the
        group of all torsion it is defined over the base field of this
        abelian variety.

        OUTPUT: a field

        EXAMPLES::

            sage: J0(23).qbar_torsion_subgroup().field_of_definition()                  # needs sage.rings.number_field
            Rational Field
        """
        return self.__abvar.base_field()

    def _element_constructor_(self, x):
        r"""
        Create an element in this torsion subgroup.

        INPUT:

        - ``x`` -- vector in `\QQ^{2d}`

        OUTPUT: torsion point

        EXAMPLES::

            sage: P = J0(23).qbar_torsion_subgroup()([1,1/2,3/4,2]); P                  # needs sage.rings.number_field
            [(1, 1/2, 3/4, 2)]
            sage: P.order()                                                             # needs sage.rings.number_field
            4
        """
        v = self.__abvar.vector_space()(x)
        return self.element_class(self, v)

    def abelian_variety(self):
        """
        Return the abelian variety that this is the set of all torsion
        points on.

        OUTPUT: abelian variety

        EXAMPLES::

            sage: J0(23).qbar_torsion_subgroup().abelian_variety()                      # needs sage.rings.number_field
            Abelian variety J0(23) of dimension 2
        """
        return self.__abvar
